NOUN_PHRASE = "noun-phrase"
NOUN = "noun"
VERB_PHRASE = "verb-phrase"
VERB = "verb"
LEXICON = "lexicon"
PROPER_NOUN = "proper-noun"
UTTERANCE = "utterance"
REQUEST = "request"
PREDICATE = "predicate"
SINGULAR = "singular"
