ACTION_INTENT = "action"
QUESTION_INTENT = "question"
ANSWER_INTENT = "answer"
ANSWER_NEGATION_INTENT = "answer_negation"
VALID_INTENT_MOVES = [ACTION_INTENT, QUESTION_INTENT]
VALID_ANSWER_MOVES = [ANSWER_INTENT, ANSWER_NEGATION_INTENT]
NEGATIVE_INTENT = "NEGATIVE"
