ENGLISH = "eng"
RASA_ENGLISH = "en"
SWEDISH = "sv"
RASA_SWEDISH = "sv"
ITALIAN = "it"
FRENCH = "fre"
DUTCH = "dut"
CHINESE = "chi"
SPANISH = "spa"
RASA_SPANISH = "es"

SUPPORTED_RASA_LANGUAGES = [
    ENGLISH,
    SWEDISH,
    SPANISH
]

PREPACKAGED_SPACY_LANGUAGES = [
    RASA_ENGLISH,
    RASA_SPANISH
]

RASA_LANGUAGE = {
    ENGLISH: RASA_ENGLISH,
    SWEDISH: RASA_SWEDISH,
    SPANISH: RASA_SPANISH
}

SUPPORTED_LANGUAGES = [ENGLISH, SWEDISH, ITALIAN, FRENCH, DUTCH, CHINESE, SPANISH]
